package org.herac.tuxguitar.io.ptb;

import java.util.Iterator;
import java.util.List;

import org.herac.tuxguitar.io.ptb.base.PTBar;
import org.herac.tuxguitar.io.ptb.base.PTBeat;
import org.herac.tuxguitar.io.ptb.base.PTComponent;
import org.herac.tuxguitar.io.ptb.base.PTGuitarIn;
import org.herac.tuxguitar.io.ptb.base.PTNote;
import org.herac.tuxguitar.io.ptb.base.PTPosition;
import org.herac.tuxguitar.io.ptb.base.PTSection;
import org.herac.tuxguitar.io.ptb.base.PTSong;
import org.herac.tuxguitar.io.ptb.base.PTSymbol;
import org.herac.tuxguitar.io.ptb.base.PTTempo;
import org.herac.tuxguitar.io.ptb.base.PTTrack;
import org.herac.tuxguitar.io.ptb.base.PTTrackInfo;
import org.herac.tuxguitar.io.ptb.helper.TrackHelper;
import org.herac.tuxguitar.song.factory.TGFactory;
import org.herac.tuxguitar.song.managers.TGSongManager;
import org.herac.tuxguitar.song.models.TGBeat;
import org.herac.tuxguitar.song.models.TGDuration;
import org.herac.tuxguitar.song.models.TGMeasure;
import org.herac.tuxguitar.song.models.TGNote;
import org.herac.tuxguitar.song.models.TGSong;
import org.herac.tuxguitar.song.models.TGString;
import org.herac.tuxguitar.song.models.TGTrack;
import org.herac.tuxguitar.song.models.effects.TGEffectBend;

public class PTSongParser {
	
	private TGSongManager manager;
	private TrackHelper helper;

	public PTSongParser(TGFactory factory){		
		this.manager = new TGSongManager(factory);
		this.helper = new TrackHelper();
	}

	public TGSong parseSong(PTSong song){
		this.manager.setSong(this.manager.newSong());
		
		this.parseTrack(song.getTrack1(),this.manager.getSong().countTracks());
		this.parseTrack(song.getTrack2(),this.manager.getSong().countTracks() + 1);
		
		return new TGSongAdjuster(this.manager).process();
	}
	
	private void parseTrack(PTTrack track,int number){
		this.helper.reset();
		
		long start = TGDuration.QUARTER_TIME;
		for( int sIndex = 0; sIndex < track.getSections().size(); sIndex ++){
			PTSection section = (PTSection) track.getSections().get(sIndex);
			section.orderPositions();
			
			//calculo el siguiente start del compas
			this.helper.getStartHelper().init(section.getNumber(),section.getStaffs());
			this.helper.getStartHelper().initVoices(start);

			//parseo las posiciones
			for( int pIndex = 0; pIndex < section.getPositions().size(); pIndex ++){
				PTPosition position = (PTPosition)section.getPositions().get(pIndex);
				parsePosition(track,position,number);
			}

			//Calculo el start para la proxima seccion
			start = this.helper.getStartHelper().getMaxStart();
		}
	}
	
	private void parsePosition(PTTrack track,PTPosition position,int fromStaff){
		for(int i = 0; i < position.getComponents().size(); i ++){
			PTComponent component = (PTComponent)position.getComponents().get(i);
			if(component instanceof PTBar){
				parseBar((PTBar)component, fromStaff);
			}
			else if(component instanceof PTSymbol){
				parseSymbol((PTSymbol)component,fromStaff);
			}
			else if(component instanceof PTGuitarIn){
				parseGuitarIn(track,(PTGuitarIn)component, fromStaff);
			}
			else if(component instanceof PTTempo){
				parseTempo((PTTempo)component);
			}			
			else if(component instanceof PTBeat){
				parseBeat((PTBeat)component,fromStaff);
			}
		}
	}
	
	private void parseBar(PTBar bar,int fromStaff){
		this.helper.getStartHelper().initVoices(this.helper.getStartHelper().getMaxStart());
		if(bar.getRepeatClose() > 0){
			if(!this.helper.getRepeatHelper().isRepeatAlternative()){
				this.helper.getRepeatHelper().setRepeatEnd(this.helper.getStartHelper().getMaxStart());
				this.doRepeat((bar.getRepeatClose() - 1),fromStaff);
			}
		}
		if(bar.isRepeatStart()){
			this.helper.getRepeatHelper().reset();
			this.helper.getRepeatHelper().setRepeatStart(this.helper.getStartHelper().getMaxStart());
		}
	}
	
	private void parseSymbol(PTSymbol symbol,int fromStaff){
		
		if(symbol.getEndNumber() == 1){
			this.helper.getRepeatHelper().setRepeatAlternative(true);
			this.helper.getRepeatHelper().setRepeatEnd(this.helper.getStartHelper().getMaxStart());
		}
		else if(symbol.getEndNumber() > 1 && this.helper.getRepeatHelper().isRepeatAlternative()){
			doRepeat(1,fromStaff);
		}
	}
	
	private void parseGuitarIn(PTTrack track,PTGuitarIn guitarIn,int fromStaff){
		if(!this.helper.getInfoHelper().isSetted(guitarIn.getStaff())){
			PTTrackInfo info = track.getInfo(guitarIn.getTrackInfo());
			if(info != null){
				TGTrack tgTrack = getTrack( (fromStaff + guitarIn.getStaff()) );
				tgTrack.setName( info.getName() );
				tgTrack.getChannel().setInstrument((short) info.getInstrument() );
				tgTrack.getChannel().setVolume((short) info.getVolume() );
				tgTrack.getChannel().setBalance((short) info.getBalance() );
				tgTrack.getStrings().clear();
				for(int i = 0; i < info.getStrings().length; i ++){
					TGString string = this.manager.getFactory().newString();
					string.setNumber( (i + 1) );
					string.setValue( info.getStrings()[i] );
					tgTrack.getStrings().add(string);
				}
				this.helper.getInfoHelper().setStaff(guitarIn.getStaff());
			}	
		}
	}
	
	private void parseTempo(PTTempo tempo){
		TGMeasure measure = getMeasure(1,this.helper.getStartHelper().getMaxStart());
		measure.getTempo().setValue( tempo.getTempo() );
		measure.getHeader().setTripletFeel(tempo.getTripletFeel());
	}
	
	private void parseBeat(PTBeat beat,int fromStaff){
		if(beat.isGrace()){
			//TODO: agrear el efecto a las notas del siguiente beat
			return;
		}
		for(int i = 0; i < beat.getMultiBarRest(); i ++){
			TGBeat tgBeat = this.manager.getFactory().newBeat();
			tgBeat.setStart(this.helper.getStartHelper().getVoiceStart(beat.getStaff(),beat.getVoice()));
			tgBeat.getDuration().setValue(beat.getDuration());
			tgBeat.getDuration().setDotted(beat.isDotted());
			tgBeat.getDuration().setDoubleDotted(beat.isDoubleDotted());
			tgBeat.getDuration().getTupleto().setTimes(beat.getTimes());
			tgBeat.getDuration().getTupleto().setEnters(beat.getEnters());

			//si no hay notas, igualmente que se cree el compas
			TGMeasure measure = getMeasure( (fromStaff + beat.getStaff()) , tgBeat.getStart() );
			measure.addBeat(tgBeat);
			
			Iterator it = beat.getNotes().iterator();
			while(it.hasNext()){
				PTNote ptNote = (PTNote)it.next();
				TGNote note = this.manager.getFactory().newNote();
				note.setString(ptNote.getString());
				note.setValue(ptNote.getValue());
				note.setTiedNote( ptNote.isTied() );
				note.getEffect().setVibrato( beat.isVibrato() );
				note.getEffect().setDeadNote( ptNote.isDead() );
				note.getEffect().setHammer( ptNote.isHammer() );
				note.getEffect().setSlide( ptNote.isSlide() );
				note.getEffect().setBend( makeBend(ptNote.getBend()));

				tgBeat.addNote(note);
			}

			this.helper.getStartHelper().setVoiceStart(beat.getStaff(),beat.getVoice(),(tgBeat.getStart() + tgBeat.getDuration().getTime()));
		}
	}
	
	private void doRepeat(int count,int fromStaff){
		long repeatStart = this.helper.getRepeatHelper().getRepeatStart();
		long repeatEnd = this.helper.getRepeatHelper().getRepeatEnd();
		if(count > 0 && repeatStart >= TGDuration.QUARTER_TIME && repeatEnd > repeatStart){
			long start = this.helper.getStartHelper().getMaxStart();
			for(int tIndex = (fromStaff - 1); tIndex < this.manager.getSong().countTracks(); tIndex++ ){
				TGTrack track = this.manager.getSong().getTrack(tIndex);
				List measures = this.manager.getTrackManager().getMeasuresBetween(track, repeatStart , repeatEnd);
				for(int repeat = 0; repeat < count;repeat ++){
					for(int mIndex = 0; mIndex < measures.size(); mIndex ++){
						TGMeasure measure = (TGMeasure)measures.get(mIndex);
						
						for(int b = 0; b < measure.countBeats(); b ++){
							TGBeat beat = measure.getBeat( b );
							if(beat.getStart() >= repeatStart && beat.getStart() < repeatEnd){
								TGBeat clone = beat.clone(this.manager.getFactory());
								clone.setStart(  clone.getStart() + ((start - repeatStart) * (repeat + 1) ) );
								getMeasure( track.getNumber(),clone.getStart()).addBeat(clone);
							}
						}
					}
				}
			}
			this.helper.getStartHelper().initVoices( this.helper.getStartHelper().getMaxStart() + ((repeatEnd - repeatStart) * count)  );
			
			//Reseteo todo.
			this.helper.getRepeatHelper().reset();
		}
	}
	
	private TGEffectBend makeBend(int value){
		if(value >= 1 && value <= 8){
			TGEffectBend bend = this.manager.getFactory().newEffectBend();
			if(value == 1){
				bend.addPoint(0,0);
				bend.addPoint(6,(TGEffectBend.SEMITONE_LENGTH * 4));
				bend.addPoint(12,(TGEffectBend.SEMITONE_LENGTH * 4));               
			}
			else if(value == 2){
				bend.addPoint(0,0);
				bend.addPoint(3,(TGEffectBend.SEMITONE_LENGTH * 4));
				bend.addPoint(6,(TGEffectBend.SEMITONE_LENGTH * 4));
				bend.addPoint(9,0);
				bend.addPoint(12,0);
			}
			else if(value == 3){	      
				bend.addPoint(0,0);
				bend.addPoint(6,(TGEffectBend.SEMITONE_LENGTH * 4));
				bend.addPoint(12,(TGEffectBend.SEMITONE_LENGTH * 4));         
			}
			else if(value == 4){	                
				bend.addPoint(0,(TGEffectBend.SEMITONE_LENGTH * 4));
				bend.addPoint(12,(TGEffectBend.SEMITONE_LENGTH * 4));
			}
			else if(value == 5){
				bend.addPoint(0,(TGEffectBend.SEMITONE_LENGTH * 4));
				bend.addPoint(4,(TGEffectBend.SEMITONE_LENGTH * 4));
				bend.addPoint(8,0);
				bend.addPoint(12,0);
			}
			else if(value == 6){
				bend.addPoint(0,8);
				bend.addPoint(12,8);
			}
			else if(value == 7){
				bend.addPoint(0,(TGEffectBend.SEMITONE_LENGTH * 4));
				bend.addPoint(4,(TGEffectBend.SEMITONE_LENGTH * 4));
				bend.addPoint(8,0);
				bend.addPoint(12,0);
			}
			else if(value == 8){            
				bend.addPoint(0,(TGEffectBend.SEMITONE_LENGTH * 4));
				bend.addPoint(4,(TGEffectBend.SEMITONE_LENGTH * 4));
				bend.addPoint(8,0);
				bend.addPoint(12,0);
			}
			return bend;
		}
		return null;
	}	
	
	private TGMeasure getMeasure(int trackNumber,long start){
		TGMeasure measure = null;
		TGTrack track = getTrack(trackNumber);
		while( (measure = this.manager.getTrackManager().getMeasureAt(track, start)) == null){
			this.manager.addNewMeasureBeforeEnd();
		}
		return measure;
	}
	
	private TGTrack getTrack(int number){
		TGTrack track = null;
		while( (track = this.manager.getTrack(number)) == null){
			track = this.manager.createTrack();
		}
		return track;
	}
}

class TGSongAdjuster{
	
	protected TGSongManager manager;
	
	public TGSongAdjuster(TGSongManager manager){
		this.manager = manager;
	}
	
	public TGSong process(){
		Iterator tracks = this.manager.getSong().getTracks();
		while(tracks.hasNext()){
			TGTrack track = (TGTrack)tracks.next();
			Iterator measures = track.getMeasures();
			while(measures.hasNext()){
				TGMeasure measure = (TGMeasure)measures.next();
				this.process(measure);
			}
		}
		return this.manager.getSong();
	}
	
	public void process(TGMeasure measure){
		this.manager.getMeasureManager().orderBeats(measure);
		
		joinBeats(measure);
	}
	
	public void joinBeats(TGMeasure measure){
		TGBeat previous = null;
		boolean finish = true;
		
		long measureStart = measure.getStart();
		long measureEnd = (measureStart + measure.getLength());
		for(int i = 0;i < measure.countBeats();i++){
			TGBeat beat = measure.getBeat( i );
			long beatStart = beat.getStart();
			long beatLength = beat.getDuration().getTime();			
			if(previous != null){
				long previousStart = previous.getStart();
				long previousLength = previous.getDuration().getTime();
				
				if(previousStart == beatStart){
					// add beat notes to previous
					for(int n = 0;n < beat.countNotes();n++){
						TGNote note = beat.getNote( n );
						previous.addNote( note );
					}

					// add beat chord to previous
					if(!previous.isChordBeat() && beat.isChordBeat()){
						previous.setChord( beat.getChord() );
					}
				
					// add beat text to previous
					if(!previous.isTextBeat() && beat.isTextBeat()){
						previous.setText( beat.getText() );
					}
				
					// set the best duration
					if(beatLength > previousLength && (beatStart + beatLength) <= measureEnd){
						beat.getDuration().copy(previous.getDuration());
					}

					measure.removeBeat(beat);
					finish = false;
					break;
				}
			
				else if(previousStart < beatStart && (previousStart + previousLength) > beatStart){
					if(beat.isRestBeat()){
						measure.removeBeat(beat);
						finish = false;
						break;
					}
					TGDuration duration = TGDuration.fromTime(this.manager.getFactory(), (beatStart - previousStart) );
					duration.copy( previous.getDuration() );
				}
			}
			if( (beatStart + beatLength) > measureEnd ){
				if(beat.isRestBeat()){
					measure.removeBeat(beat);
					finish = false;
					break;
				}
				TGDuration duration = TGDuration.fromTime(this.manager.getFactory(), (measureEnd - beatStart) );
				duration.copy( beat.getDuration() );
			}			
			previous = beat;
		}
		if(!finish){
			joinBeats(measure);
		}
	}
}